import std/json, std/jsonutils, std/os, std/strtabs, std/tables, std/times
import timeutils, uuids, zero_functional

import ./models

type
  DbRoot = ref object
    albums: TableRef[UUID, Album]
    artists: TableRef[UUID, Artist]
    bookmarks: TableRef[UUID, Bookmark]
    mediaFiles: TableRef[UUID, MediaFile]
    playlists: TableRef[UUID, Playlist]
    tags: StringTableRef

    albumsToMediaFiles: TableRef[UUID, seq[UUID]]
    artistsToMediaFiles: TableRef[UUID, seq[UUID]]
    artistsAndAlbums: seq[tuple[artistId, albumId: UUID]]
    playlistsToMediaFiles: TableRef[UUID, seq[UUID]]
    tagsAndMediaFiles: seq[tuple[tagName: string, mediaFileId: UUID]]

  WdiwtltDb* = ref object
    jsonFilePath: string
    root: DbRoot

## API Contract
proc initDb*(path: string): WdiwtltDb;
proc loadDb*(path: string): WdiwtltDb;
proc persist*(db: WdiwtltDb): void;

proc delete*(db: WdiwtltDb, a: Album);
proc delete*(db: WdiwtltDb, a: Artist);
proc delete*(db: WdiwtltDb, b: Bookmark);
proc delete*(db: WdiwtltDb, mf: MediaFile);
proc delete*(db: WdiwtltDb, p: Playlist);
proc delete*(db: WdiwtltDb, t: Tag);

proc remove*(db: WdiwtltDb, a: Album, mf: MediaFile);
proc remove*(db: WdiwtltDb, a: Artist, mf: MediaFile);
proc remove*(db: WdiwtltDb, artist: Artist, album: Album);

proc pruneStalePlaylists*(db: WdiwtltDb, ts: DateTime);
proc pruneStaleBookmarks*(db: WdiwtltDb, ts: DateTime);
proc removeEmptyAlbums*(db: WdiwtltDb): void;
proc removeEmptyArtists*(db: WdiwtltDb): void;
proc removeEmptyPlaylists*(db: WdiwtltDb): void;

## Internals
proc toJsonHook(uuid: UUID): JsonNode = %($uuid)
proc fromJsonHook(u: var UUID,node: JsonNode) =
  u = parseUUID(node.getStr)

proc toJsonHook(dt: DateTime): JsonNode = %(dt.formatIso8601)
proc fromJsonHook(dt: var DateTime, node: JsonNode) =
  dt = parseIso8601(node.getStr)

## API Implementation
proc initDb*(path: string): WdiwtltDb =
  WdiwtltDb(
    jsonFilePath: path,
    root: DbRoot(
      albums: newTable[UUID, Album](),
      artists: newTable[UUID, Artist](),
      bookmarks: newTable[UUID, Bookmark](),
      mediaFiles: newTable[UUID, MediaFile](),
      playlists: newTable[UUID, Playlist](),
      tags: newStringTable(),
      albumsToMediaFiles: newTable[UUID, seq[UUID]](),
      artistsToMediaFiles: newTable[UUID, seq[UUID]](),
      artistsAndAlbums: @[],
      playlistsToMediaFiles: newTable[UUID, seq[UUID]](),
      tagsAndMediaFiles: @[]
  ))

proc loadDb*(path: string): WdiwtltDb =
  if not fileExists(path):
    raise newException(Exception, "Unable to open database file '" & path & "'")

  let jsonRoot = parseJson(path)
  var root: DbRoot
  root.fromJson(jsonRoot)

  result = WdiwtltDb(
    jsonFilePath: path,
    root: root)

proc persist*(db: WdiwtltDb): void =
  let jsonRoot = db.root.toJson
  db.jsonFilePath.writeFile($jsonRoot)

proc delete*(db: WdiwtltDb, a: Album) =
  db.root.albums.del(a.id)
  db.root.albumsToMediaFiles.del(a.id)
  db.root.artistsAndAlbums =
    db.root.artistsAndAlbums --> filter(it.albumId != a.id)

proc delete*(db: WdiwtltDb, a: Artist) =
  db.root.artists.del(a.id)
  db.root.artistsToMediaFiles.del(a.id)
  db.root.artistsAndAlbums =
    db.root.artistsAndAlbums --> filter(it.artistId != a.id)

proc delete*(db: WdiwtltDb, b: Bookmark) = db.root.bookmarks.del(b.id)

proc delete*(db: WdiwtltDb, mf: MediaFile) =
  db.root.mediaFiles.del(mf.id)

  for albumId, mfIds in db.root.albumsToMediaFiles.pairs:
    if mfIds.contains(mf.id):
      db.remove(db.root.albums[albumId], mf)

proc delete*(db: WdiwtltDb, p: Playlist) =
  db.root.playlists.del(p.id)
  db.root.playlistsToMediaFiles.del(p.id)
  for b in db.root.bookmarks.values:
    if b.playlistId == p.id: db.delete(b)

proc delete*(db: WdiwtltDb, t: Tag) =
  db.root.tags.del(t.name)
  db.root.tagsAndMediaFiles = db.root.tagsAndMediaFiles -->
    filter(it.tagName != t.name)

proc remove*(db: WdiwtltDb, a: Album, mf: MediaFile) =
  db.root.albumsToMediaFiles[a.id] =
    db.root.albumsToMediaFiles[a.id] -->
      filter(it != mf.id)

proc remove*(db: WdiwtltDb, a: Artist, mf: MediaFile) =
  db.root.artistsToMediaFiles[a.id] =
    db.root.artistsToMediaFiles[a.id] -->
      filter(it != mf.id)

proc remove*(db: WdiwtltDb, artist: Artist, album: Album) =
  db.root.artistsAndAlbums = db.root.artistsAndAlbums -->
    filter(it.albumId != album.id or it.artistId != artist.id)

proc pruneStalePlaylists*(db: WdiwtltDb, ts: DateTime) =
  db.root.playlists.values -->
    filter(it.lastUsed > ts).foreach(db.delete(it))

proc pruneStaleBookmarks*(db: WdiwtltDb, ts: DateTime) =
  db.root.bookmarks.values -->
    filter(it.lastUsed > ts).foreach(db.delete(it))

proc removeEmptyAlbums*(db: WdiwtltDb): void =
  let emptyAlbumIds = db.root.albumsToMediaFiles.pairs -->
    filter(it[1].len == 0).map(it[0])

  for i in emptyAlbumIds:
    db.root.albums.del(i)
    db.root.albumsToMediaFiles.del(i)
    db.root.artistsAndAlbums = db.root.artistsAndAlbums -->
      filter(it.albumId != i)

proc removeEmptyArtists*(db: WdiwtltDb): void =
  let emptyArtistIds = db.root.artistsToMediaFiles.pairs -->
    filter(it[1].len == 0).map(it[0])

  for i in emptyArtistIds:
    db.root.artists.del(i)
    db.root.artistsToMediaFiles.del(i)
    db.root.artistsAndAlbums = db.root.artistsAndAlbums -->
      filter(it.artistId != i)


proc removeEmptyPlaylists*(db: WdiwtltDb): void =
  let emptyPlaylistIds = db.root.playlistsToMediaFiles.pairs -->
    filter(it[1].len == 0).map(it[0])

  for i in emptyPlaylistIds:
    db.root.playlists.del(i)
    db.root.playlistsToMediaFiles.del(i)
    let emptyBookmarkIds = db.root.bookmarks.pairs -->
      filter(it[1].playlistId == i).map(it[0])
    for j in emptyBookmarkIds: db.root.bookmarks.del(j)
